"""
Gui for info_dialog

Author: Gerd Duscher

"""
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'untitled.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets
from pyTEMlib.microscope import microscope


class UiDialog(object):
    def __init__(self, dialog):
        dialog.setObjectName('Spectrum Info')
        dialog.resize(371, 184)

        valid_float = QtGui.QDoubleValidator()
        valid_int = QtGui.QIntValidator()

        self.TEM = []
        self.TEM = microscope.get_available_microscope_names()

        self.layout = QtWidgets.QGridLayout()
        self.layout.setVerticalSpacing(2)

        self.separator1 = QtWidgets.QLabel(dialog)
        self.separator1.setAutoFillBackground(True)
        palette = self.separator1.palette()
        palette.setColor(self.separator1.backgroundRole(), QtCore.Qt.blue)
        palette.setColor(self.separator1.foregroundRole(), QtCore.Qt.white)
        self.separator1.setAlignment(QtCore.Qt.AlignCenter)
        self.separator1.setMaximumHeight(50)

        self.separator1.setPalette(palette)
        ######################################################################
        self.separator1.setText("Energy Scale")

        row = 0
        self.layout.addWidget(self.separator1, row, 0, 1, 4)
        ######################################################################

        row += 1
        self.offsetLabel = QtWidgets.QLabel("Offset")
        self.offsetEdit = QtWidgets.QLineEdit(" 1.00")
        self.offsetEdit.setValidator(valid_float)
        self.offsetUnit = QtWidgets.QLabel("eV")

        self.layout.addWidget(self.offsetLabel, row, 0)
        self.layout.addWidget(self.offsetEdit, row, 1)
        self.layout.addWidget(self.offsetUnit, row, 2)

        row += 1
        self.dispersionLabel = QtWidgets.QLabel("Dispersion")
        self.dispersionEdit = QtWidgets.QLineEdit(" 1.00")
        self.dispersionEdit.setValidator(valid_float)
        self.dispersionUnit = QtWidgets.QLabel("eV/channel")

        self.layout.addWidget(self.dispersionLabel, row, 0)
        self.layout.addWidget(self.dispersionEdit, row, 1)
        self.layout.addWidget(self.dispersionUnit, row, 2)

        row += 1
        self.energy_button = QtWidgets.QPushButton('Set Energy Scale', dialog)
        self.energy_button.setCheckable(True)

        self.layout.addWidget(self.energy_button, row, 1)

        row += 1
        self.timeLabel = QtWidgets.QLabel("Exp. Time")
        self.timeEdit = QtWidgets.QLineEdit(" 100.0")
        self.timeEdit.setValidator(valid_float)
        self.timeUnit = QtWidgets.QLabel("s")

        self.layout.addWidget(self.timeLabel, row, 0)
        self.layout.addWidget(self.timeEdit, row, 1)
        self.layout.addWidget(self.timeUnit, row, 2)

        self.separator2 = QtWidgets.QLabel(dialog)
        self.separator2.setAutoFillBackground(True)
        self.separator2.setAlignment(QtCore.Qt.AlignCenter)
        self.separator2.setMaximumHeight(50)
        self.separator2.setPalette(palette)

        row += 1
        ######################################################################
        self.separator2.setText("Microscope")
        self.layout.addWidget(self.separator2, row, 0, 1, 4)
        ######################################################################

        row += 1
        self.TEMList = QtWidgets.QComboBox()
        self.TEMList.setEditable(False)
        self.TEMList.addItems(self.TEM)

        self.layout.addWidget(self.TEMList, row, 1)

        row += 1
        self.convLabel = QtWidgets.QLabel("Conv. Angle")
        self.convEdit = QtWidgets.QLineEdit(" 100.0")
        self.convEdit.setValidator(valid_float)
        self.convUnit = QtWidgets.QLabel("mrad")
        self.layout.addWidget(self.convLabel, row, 0)
        self.layout.addWidget(self.convEdit, row, 1)
        self.layout.addWidget(self.convUnit, row, 2)

        row += 1
        self.collLabel = QtWidgets.QLabel("Coll. Angle")
        self.collEdit = QtWidgets.QLineEdit(" 10.0")
        self.collEdit.setValidator(valid_float)
        self.collUnit = QtWidgets.QLabel("mrad")
        self.layout.addWidget(self.collLabel, row, 0)
        self.layout.addWidget(self.collEdit, row, 1)
        self.layout.addWidget(self.collUnit, row, 2)

        row += 1
        self.E0Label = QtWidgets.QLabel("Acc. Voltage")
        self.E0Edit = QtWidgets.QLineEdit(" 100.0")
        self.E0Edit.setValidator(valid_float)
        self.E0Unit = QtWidgets.QLabel("kV")
        self.layout.addWidget(self.E0Label, row, 0)
        self.layout.addWidget(self.E0Edit, row, 1)
        self.layout.addWidget(self.E0Unit, row, 2)

        self.separator3 = QtWidgets.QLabel(dialog)
        self.separator3.setAutoFillBackground(True)
        self.separator3.setAlignment(QtCore.Qt.AlignCenter)
        self.separator3.setMaximumHeight(50)
        self.separator3.setPalette(palette)

        row += 1
        ######################################################################
        self.separator3.setText("Quantification")
        self.layout.addWidget(self.separator3, row, 0, 1, 4)
        ######################################################################

        row += 1
        self.binningLabel = QtWidgets.QLabel("Binning")
        self.binningEdit = QtWidgets.QLineEdit(" 1")
        self.binningEdit.setValidator(valid_int)
        self.layout.addWidget(self.binningLabel, row, 0)
        self.layout.addWidget(self.binningEdit, row, 1)

        row += 1
        self.conversionLabel = QtWidgets.QLabel("Conversion")
        self.conversionEdit = QtWidgets.QLineEdit(" 25.0")
        self.conversionEdit.setValidator(valid_float)
        self.conversionUnit = QtWidgets.QLabel("e<sup>-</sup>/counts")
        self.layout.addWidget(self.conversionLabel, row, 0)
        self.layout.addWidget(self.conversionEdit, row, 1)
        self.layout.addWidget(self.conversionUnit, row, 2)
        row += 1
        self.fluxLabel = QtWidgets.QLabel("Flux")
        self.fluxEdit = QtWidgets.QLineEdit(" 100.0")
        self.fluxEdit.setValidator(valid_float)
        self.fluxUnit = QtWidgets.QLabel("e<sup>-</sup>/s")
        self.layout.addWidget(self.fluxLabel, row, 0)
        self.layout.addWidget(self.fluxEdit, row, 1)
        self.layout.addWidget(self.fluxUnit, row, 2)
        row += 1
        self.VOALabel = QtWidgets.QLabel("Measurement")
        self.VOAEdit = QtWidgets.QLineEdit(" 10.0")
        self.VOAEdit.setValidator(valid_float)
        self.VOAUnit = QtWidgets.QLabel("pA")
        self.layout.addWidget(self.VOALabel, row, 0)
        self.layout.addWidget(self.VOAEdit, row, 1)
        self.layout.addWidget(self.VOAUnit, row, 2)

        self.separator4 = QtWidgets.QLabel(dialog)
        self.separator4.setAutoFillBackground(True)
        self.separator4.setAlignment(QtCore.Qt.AlignCenter)
        self.separator4.setMaximumHeight(50)
        self.separator4.setPalette(palette)
        ######################################################################
        self.separator4.setText("Spectrum Image")

        row += 1
        self.layout.addWidget(self.separator4, row, 0, 1, 4)
        ######################################################################

        self.binXLabel = QtWidgets.QLabel('Bin X:')
        self.binXEdit = QtWidgets.QLineEdit('1')
        self.binXEdit.setValidator(valid_int)
        self.binXUnit = QtWidgets.QLabel('pixel')

        row += 1
        self.layout.addWidget(self.binXLabel, row, 0)
        self.layout.addWidget(self.binXEdit, row, 1)
        self.layout.addWidget(self.binXUnit, row, 2)

        self.binYLabel = QtWidgets.QLabel('Bin Y:')
        self.binYEdit = QtWidgets.QLineEdit('1')
        self.binYEdit.setValidator(valid_int)
        self.binYUnit = QtWidgets.QLabel('pixel')

        row += 1
        self.layout.addWidget(self.binYLabel, row, 0)
        self.layout.addWidget(self.binYEdit, row, 1)
        self.layout.addWidget(self.binYUnit, row, 2)

        dialog.setLayout(self.layout)

        dialog.setWindowTitle("Spectrum Info")

        QtCore.QMetaObject.connectSlotsByName(dialog)



# -*- coding: utf-8 -*-
# Gleaned from
# Copyright © 2007 Francisco Javier de la Peña
# file of EELSLab.
#
"""
config_dir: setup of directory ~/.pyTEMlib for custom sources and database
"""
import os
import shutil

config_files = ['microscopes.csv', 'edges_db.csv', 'edges_db.pkl', 'fparam.txt']
data_path = os.sep.join([os.path.dirname(__file__), 'data'])

if os.name == 'posix':
    config_path = os.path.join(os.path.expanduser('~'), '.TEMlib')
    os_name = 'posix'
elif os.name in ['nt', 'dos']:
    config_path = os.path.join(os.path.expanduser('~'), '.TEMlib')
    os_name = 'windows'
else:
    config_path = '.'
if os.path.isdir(config_path) is False:
    # messages.information("Creating config directory: %s" % config_path)
    os.mkdir(config_path)

for file in config_files:
    templates_file = os.path.join(data_path, file)
    config_file = os.path.join(config_path, file)
    if os.path.isfile(config_file) is False:
        shutil.copy(templates_file, config_file)



"""
Dynamic Scattering Library for Multi-Slice Calculations

author: Gerd Duscher
"""

import numpy as np
import scipy.constants
import scipy.special

import pyTEMlib.kinematic_scattering as ks         # kinematic scattering Library


def potential_1dim(element, r):
    """ Calculates the projected potential of an atom of element

    The projected potential will be in units of V nm^2,
    however, internally we will use Angstrom instead of nm!
    The basis for this calculations are the atomic form factors of Kirkland 2𝑛𝑑 edition
    following the equation in  Appendix C page 252.

    Parameter
    ---------
    element: str
        name of 'element
    r: numpy array [nxn]
        impact parameters (distances from atom position) in nm

    Returns
    -------
    numpy array (nxn)
        projected potential in units of V nm^2
    """

    # get elementary constants
    a0 = scipy.constants.value('Bohr radius') * 1e10  # in Angstrom
    rydberg_div_e = scipy.constants.value('Rydberg constant times hc in eV')  # in V
    e0 = 2 * rydberg_div_e * scipy.constants.value('Bohr radius') * 1e10      # now in V A

    # conversion to Angstrom
    r = r * 10

    pre_factor = 2 * np.pi ** 2 * a0 * e0

    param = ks.electronFF[element]  # parametrized form factors
    f_lorentz = r * 0  # Lorentzian term
    f_gauss = r * 0  # Gaussian term
    for i in range(3):
        f_lorentz += param['fa'][i] * scipy.special.k0(2 * np.pi * r * np.sqrt(param['fb'][i]))
        f_gauss += param['fc'][i] / param['fd'][i] * np.exp(-np.pi ** 2 * r ** 2 / param['fd'][i])
    f_lorentz[0, 0] = f_lorentz[0, 1]
    # / 100 is  conversion from V Angstrom^2 to V  nm^2
    return pre_factor * (2 * f_lorentz + f_gauss) / 100  # V-nm^2


def potential_2dim(element, nx, ny, n_cell_x, n_cell_y, lattice_parameter, base):
    """Make a super-cell with potentials

    Limitation is that we only place atom potential with single pixel resolution
    """
    n_cell_x = int(2 ** np.log2(n_cell_x))
    n_cell_y = int(2 ** np.log2(n_cell_y))

    pixel_size = lattice_parameter / (nx / n_cell_x)

    a_nx = a_ny = int(1 / pixel_size)
    x, y = np.mgrid[0:a_nx, 0:a_ny] * pixel_size
    a = int(nx / n_cell_x)
    r = x ** 2 + y ** 2

    atom_potential = potential_1dim(element, r)

    potential = np.zeros([nx, ny])

    atom_potential_corner = np.zeros([nx, ny])
    atom_potential_corner[0:a_nx, 0:a_ny] = atom_potential
    atom_potential_corner[nx - a_nx:, 0:a_ny] = np.flip(atom_potential, axis=0)
    atom_potential_corner[0:a_nx, ny - a_ny:] = np.flip(atom_potential, axis=1)
    atom_potential_corner[nx - a_nx:, ny - a_ny:] = np.flip(np.flip(atom_potential, axis=0), axis=1)

    unit_cell_base = np.array(base) * a
    unit_cell_base = np.array(unit_cell_base, dtype=int)

    for pos in unit_cell_base:
        potential = potential + np.roll(atom_potential_corner, shift=np.array(pos), axis=[0, 1])

    for column in range(int(np.log2(n_cell_x))):
        potential = potential + np.roll(potential, shift=2 ** column * a, axis=1)
    for row in range(int(np.log2(n_cell_y))):
        potential = potential + np.roll(potential, shift=2 ** row * a, axis=0)

    return potential


def interaction_parameter(acceleration_voltage):
    """Calculates interaction parameter sigma

    Parameter
    ---------
    acceleration_voltage: float
        acceleration voltage in volt

    Returns
    -------
    interaction parameter: float
        interaction parameter (dimensionless)
    """
    e0 = 510998.95  # m_0 c^2 in eV

    wavelength = ks.get_wavelength(acceleration_voltage)
    e = acceleration_voltage

    return 2. * np.pi / (wavelength * e) * (e0 + e) / (2. * e0 + e)


def get_transmission(potential, acceleration_voltage):
    """ Get transmission function

    has to be multiplied in real space with wave function

    Parameter
    ---------
    potential: numpy array (nxn)
        potential of a layer
    acceleration_voltage: float
        acceleration voltage in V

    Returns
    -------
    complex numpy array (nxn)
    """

    sigma = interaction_parameter(acceleration_voltage)

    return np.exp(1j * sigma * potential)


def get_propagator(size_in_pixel, delta_z, number_layers, wavelength, field_of_view, bandwidth_factor, verbose=True):
    """Get propagator function

    has to be convoluted with wave function after transmission

    Parameter
    ---------
    size_in_pixel: int
        number of pixels of one axis in square image
    delta_z: float
        distance between layers
    number_layers: int
        number of layers to make a propagator
    wavelength: float
        wavelength of incident electrons
    field_of_view: float
        field of view of image
    bandwidth_factor: float
        relative bandwidth to avoid anti-aliasing

    Returns
    -------
    propagator: complex numpy array (layers x size_in_pixel x size_in_pixel)

    """

    k2max = size_in_pixel / field_of_view / 2. * bandwidth_factor
    print(k2max)
    if verbose:
        print(f"Bandwidth limited to a real space resolution of {1.0 / k2max * 1000} pm")
        print(f"   (= {wavelength * k2max * 1000.0:.2f} mrad)  for symmetrical anti-aliasing.")
    k2max = k2max * k2max

    kx, ky = np.mgrid[-size_in_pixel / 2:size_in_pixel / 2, -size_in_pixel / 2:size_in_pixel / 2] / field_of_view
    k_square = kx ** 2 + ky ** 2
    k_square[k_square > k2max] = 0  # bandwidth limiting

    if verbose:
        temp = np.zeros([size_in_pixel, size_in_pixel])
        temp[k_square > 0] = 1
        print(f"Number of symmetrical non-aliasing beams = {temp.sum():.0f}")

    propagator = np.zeros([number_layers, size_in_pixel, size_in_pixel], dtype=complex)
    for i in range(number_layers):
        propagator[i] = np.exp(-1j * np.pi * wavelength * k_square * delta_z[i])

    return propagator


def multi_slice(wave, number_of_unit_cell_z, number_layers, transmission, propagator):
    """Multi-Slice Calculation

    The wave function will be changed iteratively

    Parameters
    ----------
    wave: complex numpy array (nxn)
        starting wave function
    number_of_unit_cell_z: int
        this gives the thickness in multiples of c lattice parameter
    number_layers: int
        number of layers per unit cell
    transmission: complex numpy array
        transmission function
    propagator: complex numpy array
        propagator function

    Returns
    -------
    complex numpy array
    """

    for i in range(number_of_unit_cell_z):
        for layer in range(number_layers):
            wave = wave * transmission[layer]  # transmission  - real space
            wave = np.fft.fft2(wave)
            wave = wave * propagator[layer]  # propagation; propagator is defined in reciprocal space
            wave = np.fft.ifft2(wave)  # back to real space
    return wave



""" dft simulations tools

Part of pyTEMlib
by Gerd Duscher
created 10/29/2020

Supports the conversion of DFT data to simulated EELS spectra

- exciting_get_spectra:  importing dielectric function from the exciting program
- final_state_broadening: apply final state broadening to loss-spectra
"""

import numpy as np
from lxml import etree


def exciting_get_spectra(file):
    """get EELS spectra from exciting calculation"""

    tags = {'data': {}}

    tree = etree.ElementTree(file=file)
    root = tree.getroot()

    data = tags['data']

    if root.tag in ['loss', 'dielectric']:
        print(' reading ', root.tag, ' function from file ', file)
        # print(root[0].tag, root[0].text)
        map_def = root[0]
        i = 0
        v = {}
        for child_of_root in map_def:
            data[child_of_root.tag] = child_of_root.attrib
            v[child_of_root.tag] = []
            i += 1

        for elem in tree.iter(tag='map'):
            m_dict = elem.attrib
            for key in m_dict:
                v[key].append(float(m_dict[key]))

        for key in data:
            data[key]['data'] = np.array(v[key])
        data['type'] = root.tag+' function'
        return tags


def final_state_broadening(x, y, start, instrument):
    """Final state smearing of ELNES edges

    Parameters
    ----------
    x: numpy array
        x or energy loss axis of density of states
    y: numpy array
        y or intensity axis of density of states
    start: float
        start energy of edge
    instrument: float
        instrument broadening

    Return
    ------
    out_data: numpy array
        smeared intensity according to final state and instrument broadening
    """

    # Getting the smearing
    a_i = 107.25*5
    b_i = 0.04688*2.
    x = np.array(x)-start
    zero = int(-x[0]/(x[1]-x[0]))+1
    smear_i = x*0.0
    smear_i[zero:-1] = (a_i/x[zero:-1]**2)+b_i*np.sqrt(x[zero:-1])
    h_bar = 6.58e-16  # h/2pi
    pre = 1.0
    m = 6.58e-31
    smear = x*0.0
    smear[zero:-1] = pre*(h_bar/(smear_i[zero:-1]*0.000000001))*np.sqrt((2*x[zero:-1]*1.6E-19)/m)

    def lorentzian(xx, pp):
        yy = ((0.5 * pp[1]/3.14)/((xx-pp[0])**2 + ((pp[1]/2)**2)))
        return yy/sum(yy)

    p = [0, instrument]
    in_data = y.copy()
    out_data = np.array(y)*0.0
    for i in range(zero+5, len(x)):
        p[0] = x[i]
        p[1] = smear[i]/1.0
        lor = lorentzian(x+1e-9, p)
        out_data[i] = sum(in_data*lor)
        if np.isnan(out_data[i]):
            out_data[i] = 0.0

    p[1] = instrument
    in_data = out_data.copy()
    for i in range(zero-5, len(x)):
        p[0] = x[i]
        lor = lorentzian(x+1e-9, p)
        out_data[i] = sum(in_data*lor)
        # print(out_data[i],in_data[i], lor[i],in_data[i-1], lor[i-1], )
    return out_data



"""GUI definitions for peak_fit_dialog"""

# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'untitled.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class UiDialog(object):
    """GUI definitions for peak_fit_dialog"""

    def __init__(self, dialog):
        dialog.setObjectName('Fit Peaks')
        dialog.resize(371, 184)

        valid_float = QtGui.QDoubleValidator()
        valid_int = QtGui.QIntValidator()
        valid_int.setBottom(0)

        self.layout = QtWidgets.QGridLayout()
        self.layout.setVerticalSpacing(2)

        self.separator1 = QtWidgets.QLabel(dialog)
        self.separator1.setAutoFillBackground(True)
        palette = self.separator1.palette()
        palette.setColor(self.separator1.backgroundRole(), QtCore.Qt.blue)
        palette.setColor(self.separator1.foregroundRole(), QtCore.Qt.white)
        self.separator1.setAlignment(QtCore.Qt.AlignCenter)
        self.separator1.setMaximumHeight(50)

        self.separator1.setPalette(palette)
        row = 0
        ######################################################################
        self.separator1.setText("Fit Region")
        self.layout.addWidget(self.separator1, row, 0, 1, 4)
        ######################################################################

        row += 1
        name = 'fit_start'
        self.label1 = QtWidgets.QLabel('Fit Start:')
        self.edit1 = QtWidgets.QLineEdit('0')
        self.edit1.setValidator(valid_float)
        self.unit1 = QtWidgets.QLabel('eV')
        self.edit1.setObjectName(name+'_edit')
        self.unit1.setObjectName(name+'_units')

        self.layout.addWidget(self.label1, row, 0)
        self.layout.addWidget(self.edit1, row, 1)
        self.layout.addWidget(self.unit1, row, 2)

        row += 1
        name = 'fit_end'
        self.label2 = QtWidgets.QLabel('Fit End:')
        self.edit2 = QtWidgets.QLineEdit('0')
        self.edit2.setValidator(valid_float)
        self.unit2 = QtWidgets.QLabel('eV')
        self.edit2.setObjectName(name + '_edit')
        self.unit2.setObjectName(name + '_units')

        self.layout.addWidget(self.label2, row, 0)
        self.layout.addWidget(self.edit2, row, 1)
        self.layout.addWidget(self.unit2, row, 2)

        self.separator2 = QtWidgets.QLabel(dialog)
        self.separator2.setAutoFillBackground(True)
        self.separator2.setAlignment(QtCore.Qt.AlignCenter)
        self.separator2.setMaximumHeight(50)
        self.separator2.setPalette(palette)

        row += 1
        ######################################################################
        self.separator2.setText("Processing")
        self.layout.addWidget(self.separator2, row, 0, 1, 4)
        ######################################################################
        row += 1
        # self.progress = QtWidgets.QProgressBar()
        # self.progress.setGeometry(0, 0, 300, 25)
        # self.progress.setMaximum(100)

        self.smooth_label = QtWidgets.QLabel('Iterations:')
        self.iteration_list = ['0', '1', '2', '3', '4']
        self.smooth_list = QtWidgets.QComboBox()
        self.smooth_list.setEditable(False)
        self.smooth_list.addItems(self.iteration_list)
        self.smooth_list.setCurrentIndex(1)

        self.smooth_button = QtWidgets.QPushButton('Smooth', dialog)
        self.smooth_button.setCheckable(False)

        self.layout.addWidget(self.smooth_label, row, 0)
        self.layout.addWidget(self.smooth_list, row, 1)
        self.layout.addWidget(self.smooth_button, row, 2)

        row += 1
        self.find_label = QtWidgets.QLabel('Nr. of Peaks:')
        self.find_edit = QtWidgets.QLineEdit('1')
        self.find_edit.setValidator(valid_int)
        self.find_button = QtWidgets.QPushButton('Find Peaks', dialog)
        self.find_button.setCheckable(False)

        self.layout.addWidget(self.find_label, row, 0)
        self.layout.addWidget(self.find_edit, row, 1)
        self.layout.addWidget(self.find_button, row, 2)

        self.separator2 = QtWidgets.QLabel(dialog)
        self.separator2.setAutoFillBackground(True)
        self.separator2.setAlignment(QtCore.Qt.AlignCenter)
        self.separator2.setMaximumHeight(50)
        self.separator2.setPalette(palette)

        row += 1
        ######################################################################
        self.separator2.setText("Peaks")
        self.layout.addWidget(self.separator2, row, 0, 1, 4)
        ######################################################################

        row += 1
        name = 'peak'
        self.label3 = QtWidgets.QLabel('Peaks:')
        self.peak_list = ['Peak 1', 'Add Peak']
        self.list3 = QtWidgets.QComboBox()
        self.list3.setEditable(False)
        self.list3.addItems(self.peak_list)
        self.unit3 = QtWidgets.QLabel('')

        self.layout.addWidget(self.label3, row, 0)
        self.layout.addWidget(self.list3, row, 1)
        self.layout.addWidget(self.unit3, row, 2)

        self.list3.setObjectName(name + '_list')
        # self.check3.setObjectName(name + '_check')

        name = 'symmetry'
        self.edge_sym = ['Gauss', 'Lorentzian', 'Drude', 'ZL']

        self.label4 = QtWidgets.QLabel('Symmetry:')
        self.list4 = QtWidgets.QComboBox()
        self.list4.setEditable(False)
        self.list4.addItems(self.edge_sym)

        row += 1
        self.layout.addWidget(self.label4, row, 0)
        self.layout.addWidget(self.list4, row, 1)

        self.list4.setObjectName(name + '_list')

        name = 'position'
        self.label5 = QtWidgets.QLabel('Position:')
        self.edit5 = QtWidgets.QLineEdit('1')
        self.unit5 = QtWidgets.QLabel('eV')
        self.edit5.setObjectName(name + '_edit')

        row += 1
        self.layout.addWidget(self.label5, row, 0)
        self.layout.addWidget(self.edit5, row, 1)
        self.layout.addWidget(self.unit5, row, 2)

        row += 1
        name = 'amplitude'
        self.label6 = QtWidgets.QLabel('Amplitude:')
        self.edit6 = QtWidgets.QLineEdit('0')
        self.edit6.setValidator(valid_float)
        self.unit6 = QtWidgets.QLabel('counts')
        self.edit6.setObjectName(name + '_edit')
        self.unit6.setObjectName(name + '_units')

        self.layout.addWidget(self.label6, row, 0)
        self.layout.addWidget(self.edit6, row, 1)
        self.layout.addWidget(self.unit6, row, 2)

        name = 'width'
        self.label7 = QtWidgets.QLabel('Width (FWHM):')
        self.edit7 = QtWidgets.QLineEdit('0')
        self.edit7.setValidator(valid_float)
        self.unit7 = QtWidgets.QLabel('eV')
        self.edit7.setObjectName(name + '_edit')
        self.unit7.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label7, row, 0)
        self.layout.addWidget(self.edit7, row, 1)
        self.layout.addWidget(self.unit7, row, 2)

        row += 1
        name = 'asymmetry'
        self.label8 = QtWidgets.QLabel('Asymmetry:')
        self.edit8 = QtWidgets.QLineEdit('0')
        self.edit8.setValidator(valid_float)
        self.unit8 = QtWidgets.QLabel('%')
        self.label8.setObjectName(name + '_name')
        self.edit8.setObjectName(name + '_edit')
        self.unit8.setObjectName(name + '_units')

        self.layout.addWidget(self.label8, row, 0)
        self.layout.addWidget(self.edit8, row, 1)
        self.layout.addWidget(self.unit8, row, 2)

        row += 1
        self.fit_button = QtWidgets.QPushButton('Fit Peaks', dialog)
        self.fit_button.setCheckable(False)

        self.layout.addWidget(self.fit_button, row, 2)

        row += 1
        self.separator3 = QtWidgets.QLabel(dialog)
        self.separator3.setAutoFillBackground(True)
        self.separator3.setAlignment(QtCore.Qt.AlignCenter)
        self.separator3.setMaximumHeight(50)
        self.separator3.setPalette(palette)

        ######################################################################
        self.separator3.setText("Analysis")
        self.layout.addWidget(self.separator3, row, 0, 1, 4)
        ######################################################################

        row += 1
        self.labelwl = QtWidgets.QLabel('White-Line Ratio:')
        self.wl_list = ['Ratio']
        self.listwl = QtWidgets.QComboBox()
        self.listwl.setEditable(False)
        self.listwl.addItems(self.wl_list)
        self.unitswl = QtWidgets.QLabel('')

        self.layout.addWidget(self.labelwl, row, 0)
        self.layout.addWidget(self.listwl, row, 1)
        self.layout.addWidget(self.unitswl, row, 2)

        row += 1
        self.labelwls = QtWidgets.QLabel('White-Line Sum:')
        self.wls_list = ['Sum']
        self.listwls = QtWidgets.QComboBox()
        self.listwls.setEditable(False)
        self.listwls.addItems(self.wls_list)
        self.unitswls = QtWidgets.QLabel('')

        self.layout.addWidget(self.labelwls, row, 0)
        self.layout.addWidget(self.listwls, row, 1)
        self.layout.addWidget(self.unitswls, row, 2)

        dialog.setLayout(self.layout)
        dialog.setWindowTitle('Peak Fit Input')

        # self.buttonBox.accepted.connect(Dialog.accept)
        # self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(dialog)



""" Atom detection

All atom detection is done here
Everything is in unit of pixel!!

Author: Gerd Duscher

part of pyTEMlib

a pycroscopy package
"""

import numpy as np
import sys

# from skimage.feature import peak_local_max
from skimage.feature import blob_log

from sklearn.cluster import KMeans
from scipy.spatial import cKDTree
import scipy.optimize as optimization

import pyTEMlib.probe_tools as probe_tools
import pyTEMlib.file_tools as ft
import sidpy


def find_atoms(image, atom_size=0.1, threshold=0.):
    """ Find atoms is a simple wrapper for blob_log in skimage.feature

    Parameters
    ----------
    image: sidpy.Dataset
        the image to find atoms
    atom_size: float
        visible size of atom blob diameter in nm gives minimal distance between found blobs
    threshold: float
        threshold for blob finder; (usually between 0.001 and 1.0) for threshold <= 0 we use the RMS contrast

    Returns
    -------
    atoms: numpy array(Nx3)
        atoms positions and radius of blob
    """

    if not isinstance(image, sidpy.Dataset):
        raise TypeError('We need a sidpy.Dataset')
    if image.data_type.name != 'IMAGE':
        raise TypeError('We need sidpy.Dataset of sidpy.Datatype: IMAGE')
    if not isinstance(atom_size, (float, int)):
        raise TypeError('atom_size parameter has to be a number')
    if not isinstance(threshold, float):
        raise TypeError('threshold parameter has to be a float number')

    scale_x = ft.get_slope(image.dim_0)
    im = np.array(image-image.min())
    im = im/im.max()
    if threshold <= 0.:
        threshold = np.std(im)
    atoms = blob_log(im, max_sigma=atom_size/scale_x, threshold=threshold)

    return atoms


def atoms_clustering(atoms, mid_atoms, number_of_clusters=3, nearest_neighbours=7):
    """ A wrapper for sklearn.cluster kmeans clustering of atoms.

    Parameters
    ----------
    atoms: list or np.array (Nx2)
        list of all atoms
    mid_atoms: list or np.array (Nx2)
        atoms to be evaluated
    number_of_clusters: int
        number of clusters to sort (ini=3)
    nearest_neighbours: int
        number of nearest neighbours evaluated

    Returns
    -------
    clusters, distances, indices: numpy arrays
    """

    # get distances
    nn_tree = cKDTree(np.array(atoms)[:, 0:2])

    distances, indices = nn_tree.query(np.array(mid_atoms)[:, 0:2], nearest_neighbours)

    # Clustering
    k_means = KMeans(n_clusters=number_of_clusters, random_state=0)  # Fixing the RNG in kmeans
    k_means.fit(distances)
    clusters = k_means.predict(distances)

    return clusters, distances, indices


def gauss_difference(params, area):
    """
    Difference between part of an image and a Gaussian
    This function is used int he atom refine function of pyTEMlib

    Parameters
    ----------
    params: list
        list of Gaussian parameters [width, position_x, position_y, intensity]
    area:  numpy array
        2D matrix = part of an image

    Returns
    -------
    numpy array: flattened array of difference

    """
    gauss = probe_tools.make_gauss(area.shape[0], area.shape[1], width=params[0], x0=params[1], y0=params[2],
                                   intensity=params[3])
    return (area - gauss).flatten()


def atom_refine(image, atoms, radius, max_int=0, min_int=0, max_dist=4):
    """Fits a Gaussian in a blob of an image

    Parameters
    ----------
    image: np.array or sidpy Dataset
    atoms: list or np.array
        positions of atoms
    radius: float
        radius of circular mask to define fitting of Gaussian
    max_int: float
        optional - maximum intensity to be considered for fitting (to exclude contaminated areas for example)
    min_int: float
        optional - minimum intensity to be considered for fitting (to exclude contaminated holes for example)
    max_dist: float
        optional - maximum distance of movement of Gaussian during fitting

    Returns
    -------
    sym: dict
        dictionary containing new atom positions and other output such as intensity of the fitted Gaussian
    """
    rr = int(radius + 0.5)  # atom radius
    print('using radius ', rr, 'pixels')

    pixels = np.linspace(0, 2 * rr, 2 * rr + 1) - rr
    x, y = np.meshgrid(pixels, pixels)
    mask = (x ** 2 + y ** 2) < rr ** 2

    guess = [rr * 2, 0.0, 0.0, 1]

    sym = {'number_of_atoms': len(atoms)}

    volume = []
    position = []
    intensities = []
    maximum_area = []
    new_atoms = []
    gauss_width = []
    gauss_amplitude = []
    gauss_intensity = []
    if ft.QT_available:
        progress = ft.ProgressDialog("Refine Atom Positions")

    done = 0
    for i in range(len(atoms)):
        if ft.QT_available:
            progress.set_value(i)
        else:
            if done < int((i + 1) / len(atoms) * 50):
                done = int((i + 1) / len(atoms) * 50)
                sys.stdout.write('\r')
                # progress output :
                sys.stdout.write("[%-50s] %d%%" % ('=' * done, 2 * done))
                sys.stdout.flush()

        x, y = atoms[i][0:2]
        x = int(x)
        y = int(y)

        area = image[x - rr:x + rr + 1, y - rr:y + rr + 1]

        append = False

        if (x - rr) < 0 or y - rr < 0 or x + rr + 1 > image.shape[0] or y + rr + 1 > image.shape[1]:
            position.append(-1)
            intensities.append(-1.)
            maximum_area.append(-1.)
        else:
            position.append(1)
            intensities.append((area * mask).sum())
            maximum_area.append((area * mask).max())

        if max_int > 0:
            if area.sum() < max_int:
                if area.sum() > min_int:
                    append = True
        elif area.sum() > min_int:
            append = True

        pout = [0, 0, 0, 0]
        if append:
            if (x - rr) < 0 or y - rr < 0 or x + rr + 1 > image.shape[0] or y + rr + 1 > image.shape[1]:
                pass
            else:
                [pout, _] = optimization.leastsq(gauss_difference, guess, args=area)

            if (abs(pout[1]) > max_dist) or (abs(pout[2]) > max_dist):
                pout = [0, 0, 0, 0]

        volume.append(2 * np.pi * pout[3] * pout[0] * pout[0])

        new_atoms.append([x + pout[1], y + pout[2]])  # ,pout[0],  volume)) #,pout[3]))
        if all(v == 0 for v in pout):
            gauss_intensity.append(0.)
        else:
            gauss = probe_tools.make_gauss(area.shape[0], area.shape[1], width=pout[0], x0=pout[1], y0=pout[2],
                                           intensity=pout[3])
            gauss_intensity.append((gauss * mask).sum())
        gauss_width.append(pout[0])
        gauss_amplitude.append(pout[3])

    if ft.QT_available:
        progress.close()
    sym['inside'] = position
    sym['intensity_area'] = intensities
    sym['maximum_area'] = maximum_area
    sym['atoms'] = new_atoms
    sym['gauss_width'] = gauss_width
    sym['gauss_amplitude'] = gauss_amplitude
    sym['gauss_intensity'] = gauss_intensity
    sym['gauss_volume'] = volume

    return sym


def intensity_area(image, atoms, radius):
    """
    integrated intensity of atoms in an image with a mask around each atom of radius radius
    """
    rr = int(radius + 0.5)  # atom radius
    print('using radius ', rr, 'pixels')

    pixels = np.linspace(0, 2 * rr, 2 * rr + 1) - rr
    x, y = np.meshgrid(pixels, pixels)
    mask = np.array((x ** 2 + y ** 2) < rr ** 2)
    intensities = []
    for i in range(len(atoms)):
        x = int(atoms[i][1])
        y = int(atoms[i][0])
        area = image[x - rr:x + rr + 1, y - rr:y + rr + 1]
        if area.shape == mask.shape:
            intensities.append((area * mask).sum())
        else:
            intensities.append(-1)
    return intensities



"""
Input Dialog for Image Analysis

Author: Gerd Duscher

"""
# -*- coding: utf-8 -*-

import numpy as np
import sidpy

from PyQt5 import QtCore,  QtWidgets

from matplotlib.widgets import SpanSelector
from skimage import exposure

from pyTEMlib.image_dlg import *
from pyTEMlib.microscope import microscope

_version = 000


class ImageDialog(QtWidgets.QDialog):
    """
        Input Dialog for Image Analysis

        Opens a PyQt5 GUi Dialog that allows to set the experimental parameter necessary for a Quantification.


        The dialog operates on a sidpy dataset
        """

    def __init__(self, dataset, parent=None):
        super(ImageDialog, self).__init__(parent)
        if not isinstance(dataset, sidpy.Dataset):
            raise TypeError("we need a sidpy.Dataset")
        self.parent = parent
        self.debug = 0
        self.dataset = dataset
        self.image = np.array(self.dataset)
        self.v_min = np.array(dataset).min()
        self.v_max = np.array(dataset).max()

        self.ui = UiDialog(self)
        self.setWindowTitle('Image Info')

        self.dataset.plot()
        self.histogram()
        self.cid = self.ui.histogram.axes.figure.canvas.mpl_connect('button_press_event', self.onclick)

        self.span = SpanSelector(self.ui.histogram.axes, self.on_select, 'horizontal', useblit=False,
                                 button=1, minspan=5,
                                 rectprops=dict(alpha=0.3, facecolor='blue'))
        minimum_info = {'size': self.dataset.shape,
                        'exposure_time': 0.0,
                        'convergence_angle': 0.0,
                        'acceleration_voltage': 100.0,
                        'binning': 1, 'conversion': 1.0,
                        'flux': 1.0, 'current': 1.0}

        if 'experiment' not in self.dataset.metadata:
            self.dataset.metadata['experiment'] = {}
        for key, value in minimum_info.items():
            if key not in self.dataset.metadata['experiment']:
                self.dataset.metadata['experiment'][key] = value
        self.experiment = self.dataset.metadata['experiment']
        self.set_action()
        self.update()

    def histogram(self, bins=256):
        ax_hist = self.ui.histogram.axes
        ax_hist.clear()

        hist, bin_edges = np.histogram(np.array(self.image), range=[self.v_min, self.v_max], bins=bins, density=True)
        ax_hist.plot(np.array(bin_edges)[:-1], np.array(hist))

        image = self.image * 1.0
        image[image < self.v_min] = self.v_min
        image[image > self.v_max] = self.v_max

        img_cdf, bins = exposure.cumulative_distribution(np.array(image), bins)
        ax_hist.plot(bins, img_cdf * hist.max(), 'r')
        ax_hist.figure.canvas.draw()
        self.span = SpanSelector(self.ui.histogram.axes, self.on_select, 'horizontal', useblit=False,
                                 button=1, minspan=5,
                                 rectprops=dict(alpha=0.3, facecolor='blue'))
        self.plot()

    def onclick(self, event):
        if event.dblclick:
            self.v_min = np.array(self.dataset).min()
            self.v_max = np.array(self.dataset).max()
            self.histogram()

    def on_select(self, v_min, v_max):
        self.v_min = v_min
        self.v_max = v_max
        self.histogram()

    def plot(self):
        ax = self.dataset.view.axis
        img = self.dataset.view.img
        img.set_data(self.image)
        img.set_clim(vmin=self.v_min, vmax=self.v_max)
        ax.figure.canvas.draw()

    def on_enter(self):
        sender = self.sender()
        if sender == self.ui.timeEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['exposure_time'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.convEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['convergence_angle'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.E0Edit:
            value = float(str(sender.displayText()).strip())
            self.experiment['acceleration_voltage'] = value * 1000.0
            sender.setText(f"{value:.2f}")

    def on_list_enter(self):
        sender = self.sender()
        if sender == self.ui.TEMList:
            microscope.set_microscope(self.ui.TEMList.currentText())
            self.setWindowTitle(microscope.name)

            self.experiment['microscope'] = microscope.name
            self.experiment['convergence_angle'] = microscope.alpha
            self.experiment['acceleration_voltage'] = microscope.E0
            self.update()

    def update(self):
        self.ui.convEdit.setText(f"{self.experiment['convergence_angle']:.2f}")
        self.ui.E0Edit.setText(f"{self.experiment['acceleration_voltage']/1000.:.2f}")

        self.ui.timeEdit.setText(f"{self.experiment['exposure_time']:.6f}")
        size_text = f'{self.dataset.shape[0]}'
        for size in self.dataset.shape[1:]:
            size_text = size_text + f' x {size}'
        self.ui.sizeEdit.setText(size_text)

        # self.ui.binningEdit.setText(f"{self.experiment['binning']}")
        # self.ui.conversionEdit.setText(f"{self.experiment['conversion']:.2f}")
        # self.ui.fluxEdit.setText(f"{self.experiment['flux']:.2f}")
        # self.ui.VOAEdit.setText(f"{self.experiment['current']:.2f}")

    def set_action(self):
        self.ui.timeEdit.editingFinished.connect(self.on_enter)
        self.ui.TEMList.activated[str].connect(self.on_list_enter)

        self.ui.convEdit.editingFinished.connect(self.on_enter)
        self.ui.E0Edit.editingFinished.connect(self.on_enter)



""" default microscope parameters from config file

Read microscope CSV file

for pyTEMLib by Gerd

copyright 2012, Gerd Duscher
updated 2021
"""
# -*- coding: utf-8 -*-

import csv
import os.path

from pyTEMlib.config_dir import config_path
microscopes_file = os.path.join(config_path, 'microscopes.csv')


class Microscope(object):
    """Class to read configuration file and provide microscope information"""
    microscopes = {}
    name = None
    E0 = None
    alpha = None
    beta = None
    pppc = None
    correlation_factor = None

    def __init__(self):
        self.load_microscopes()
        default_tem = self.microscopes[list(self.microscopes.keys())[0]]
        self.set_microscope(default_tem['Microscope'])

    def load_microscopes(self):
        f = open(microscopes_file, 'r')

        labels = f.readline().strip().split(',')
        # print labels
        csv_read = csv.DictReader(f, labels, delimiter=",")

        for line in csv_read:
            tem = line['Microscope']
            self.microscopes[tem] = line
            for i in self.microscopes[tem]:
                if i != 'Microscope':
                    self.microscopes[tem][i] = float(self.microscopes[tem][i])
        f.close()

    def get_available_microscope_names(self):
        tem = []
        for scope in self.microscopes.keys():
            tem.append(scope)
        return tem

    def set_microscope(self, microscope_name):
        if microscope_name in self.microscopes:
            self.name = microscope_name


microscope = Microscope()



"""
QT dialog window for EELS compositional analysis

Author: Gerd Duscher
"""
from PyQt5 import QtCore,  QtWidgets
from pyTEMlib import eels_dlg

import numpy as np

from pyTEMlib import eels_tools as eels
from pyTEMlib import interactive_eels
import matplotlib.pylab as plt
import matplotlib.patches as patches

from pyTEMlib import file_tools as ft

import sidpy

_version = 000


class EELSDialog(QtWidgets.QDialog):
    """
    EELS Input Dialog for Chemical Analysis


    """

    def __init__(self, dataset=None):
        super().__init__(None, QtCore.Qt.WindowStaysOnTopHint)
        # Create an instance of the GUI
        if dataset is None:
            # make a dummy dataset
            dataset = ft.make_dummy_dataset(sidpy.DataType.SPECTRUM)
        if not isinstance(dataset, sidpy.Dataset):
            raise TypeError('dataset has to be a sidpy dataset')
        self.spec_dim = ft.get_dimensions_by_type('spectral', dataset)
        if len(self.spec_dim) != 1:
            raise TypeError('We need exactly one SPECTRAL dimension')
        self.spec_dim = self.spec_dim[0]

        self.ui = eels_dlg.UiDialog(self)
        # Run the .setup_ui() method to show the GUI
        # self.ui.setup_ui(self)

        self.set_action()

        self.dataset = dataset
        self.energy_scale = np.array([])
        self.model = np.array([])
        self.edges = {}

        self.show_regions = False
        self.show()

        self.set_dataset(dataset)
        # TODO: set elements does not work correctly for periodic table
        # selected_edges = eels.find_edges(dataset, sensitivity=3)
        selected_edges = []
        initial_elements = []

        for edge in selected_edges:
            initial_elements.append(edge.split('-')[0])
        if len(initial_elements) > 0:
            self.set_elements(initial_elements)

        self.pt_dialog = interactive_eels.PeriodicTableDialog(energy_scale=self.energy_scale,
                                                              initial_elements=initial_elements)
        self.pt_dialog.signal_selected[list].connect(self.set_elements)

        self.dataset.plot()

        if hasattr(self.dataset.view, 'axes'):
            self.axis = self.dataset.view.axes[-1]
        elif hasattr(self.dataset.view, 'axis'):
            self.axis = self.dataset.view.axis

        self.figure = self.axis.figure
        self.updY = 0
        self.figure.canvas.mpl_connect('button_press_event', self.plot)

        self.plot()
        self.ui.edit4.setFocus()

    def set_dataset(self, dataset):

        self.dataset = dataset
        if 'edges' not in self.dataset.metadata or self.dataset.metadata['edges'] == {}:
            self.dataset.metadata['edges'] = {'0': {}, 'model': {}}
        self.edges = self.dataset.metadata['edges']

        spec_dim = ft.get_dimensions_by_type('spectral', dataset)[0]

        if len(spec_dim) == 0:
            raise TypeError('We need at least one SPECTRAL dimension')

        self.spec_dim = spec_dim[0]
        self.energy_scale = dataset._axes[self.spec_dim].values
        self.ui.edit2.setText(f"{self.energy_scale[-2]:.3f}")

        if 'fit_area' not in self.edges:
            self.edges['fit_area'] = {}
        if 'fit_start' not in self.edges['fit_area']:
            self.ui.edit1.setText(f"{self.energy_scale[50]:.3f}")
            self.edges['fit_area']['fit_start'] = float(self.ui.edit1.displayText())
        else:
            self.ui.edit1.setText(f"{self.edges['fit_area']['fit_start']:.3f}")
        if 'fit_end' not in self.edges['fit_area']:
            self.ui.edit2.setText(f"{self.energy_scale[-2]:.3f}")
            self.edges['fit_area']['fit_end'] = float(self.ui.edit2.displayText())
        else:
            self.ui.edit2.setText(f"{self.edges['fit_area']['fit_end']:.3f}")

        if self.dataset.data_type.name == 'SPECTRAL_IMAGE':
            if 'SI_bin_x' not in self.dataset.metadata['experiment']:
                self.dataset.metadata['experiment']['SI_bin_x'] = 1
                self.dataset.metadata['experiment']['SI_bin_y'] = 1

            bin_x = self.dataset.metadata['experiment']['SI_bin_x']
            bin_y = self.dataset.metadata['experiment']['SI_bin_y']
            self.dataset.view.set_bin([bin_x, bin_y])
        self.update()

    def update(self):
        index = self.ui.list3.currentIndex()  # which edge
        edge = self.edges[str(index)]

        if 'Z' in edge:
            self.ui.list5.setCurrentIndex(self.ui.edge_sym.index(edge['symmetry']))
            self.ui.edit4.setText(str(edge['Z']))
            self.ui.unit4.setText(edge['element'])
            self.ui.edit6.setText(f"{edge['onset']:.2f}")
            self.ui.edit7.setText(f"{edge['start_exclude']:.2f}")
            self.ui.edit8.setText(f"{edge['end_exclude']:.2f}")
            self.ui.edit9.setText(f"{edge['areal_density']:.2e}")
        else:
            self.ui.list3.setCurrentIndex(0)
            self.ui.edit4.setText(str(0))
            self.ui.unit4.setText(' ')
            self.ui.edit6.setText(f"{0:.2f}")
            self.ui.edit7.setText(f"{0:.2f}")
            self.ui.edit8.setText(f"{0:.2f}")
            self.ui.edit9.setText(f"{0:.2e}")

    def update_element(self, z):
        # We check whether this element is already in the
        zz = eels.get_z(z)
        for key, edge in self.edges.items():
            if 'Z' in edge:
                if zz == edge['Z']:
                    return False

        major_edge = ''
        minor_edge = ''
        all_edges = {}
        x_section = eels.get_x_sections(zz)
        edge_start = 10  # int(15./ft.get_slope(self.energy_scale)+0.5)
        for key in x_section:
            if len(key) == 2 and key[0] in ['K', 'L', 'M', 'N', 'O'] and key[1].isdigit():
                if self.energy_scale[edge_start] < x_section[key]['onset'] < self.energy_scale[-edge_start]:
                    if key in ['K1', 'L3', 'M5']:
                        major_edge = key
                    elif key in self.ui.edge_sym:
                        if minor_edge == '':
                            minor_edge = key
                        if int(key[-1]) % 2 > 0:
                            if int(minor_edge[-1]) % 2 == 0 or key[-1] > minor_edge[-1]:
                                minor_edge = key

                    all_edges[key] = {'onset': x_section[key]['onset']}

        if major_edge != '':
            key = major_edge
        elif minor_edge != '':
            key = minor_edge
        else:
            print(f'Could not find no edge of {zz} in spectrum')
            return False

        index = self.ui.list3.currentIndex()

        if str(index) not in self.edges:
            self.edges[str(index)] = {}

        start_exclude = x_section[key]['onset'] - x_section[key]['excl before']
        end_exclude = x_section[key]['onset'] + x_section[key]['excl after']

        self.edges[str(index)] = {'Z': zz, 'symmetry': key, 'element': eels.elements[zz],
                                  'onset': x_section[key]['onset'], 'end_exclude': end_exclude,
                                  'start_exclude': start_exclude}
        self.edges[str(index)]['all_edges'] = all_edges
        self.edges[str(index)]['chemical_shift'] = 0.0
        self.edges[str(index)]['areal_density'] = 0.0
        self.edges[str(index)]['original_onset'] = self.edges[str(index)]['onset']
        return True

    def on_enter(self):
        sender = self.sender()
        edge_list = self.ui.list3

        if sender.objectName() == 'fit_start_edit':
            value = float(str(sender.displayText()).strip())
            if value < self.energy_scale[0]:
                value = self.energy_scale[0]
            if value > self.energy_scale[-5]:
                value = self.energy_scale[-5]
            self.edges['fit_area']['fit_start'] = value
            sender.setText(str(self.edges['fit_area']['fit_start']))
        elif sender.objectName() == 'fit_end_edit':
            value = float(str(sender.displayText()).strip())
            if value < self.energy_scale[5]:
                value = self.energy_scale[5]
            if value > self.energy_scale[-1]:
                value = self.energy_scale[-1]
            self.edges['fit_area']['fit_end'] = value
            sender.setText(str(self.edges['fit_area']['fit_end']))
        elif sender.objectName() == 'element_edit':
            if str(sender.displayText()).strip() == '0':
                sender.setText('PT')
                self.pt_dialog.energy_scale = self.energy_scale
                self.pt_dialog.show()
            else:
                self.update_element(str(sender.displayText()).strip())
            self.update()
        elif sender.objectName() in ['onset_edit', 'excl_start_edit', 'excl_end_edit']:
            self.check_area_consistency()

        elif sender.objectName() == 'multiplier_edit':
            index = edge_list.currentIndex()
            self.edges[str(index)]['areal_density'] = float(self.ui.edit9.displayText())
            if 'background' not in self.edges['model']:
                print(' no background')
                return
            self.model = self.edges['model']['background']
            for key in self.edges:
                if key.isdigit():
                    self.model = self.model + self.edges[key]['areal_density'] * self.edges[key]['data']
            self.plot()
        else:
            return
        if self.show_regions:
            self.plot()

    def sort_elements(self):
        onsets = []
        for index, edge in self.edges.items():
            if index.isdigit():
                onsets.append(float(edge['onset']))

        arg_sorted = np.argsort(onsets)
        edges = self.edges.copy()
        for index, i_sorted in enumerate(arg_sorted):
            self.edges[str(index)] = edges[str(i_sorted)].copy()

        index = 0
        edge = self.edges['0']
        dispersion = self.energy_scale[1]-self.energy_scale[0]

        while str(index + 1) in self.edges:
            next_edge = self.edges[str(index + 1)]
            if edge['end_exclude'] > next_edge['start_exclude'] - 5 * dispersion:
                edge['end_exclude'] = next_edge['start_exclude'] - 5 * dispersion
            edge = next_edge
            index += 1

        if edge['end_exclude'] > self.energy_scale[-3]:
            edge['end_exclude'] = self.energy_scale[-3]

    def set_elements(self, selected_elements):

        edge_list = self.ui.list3
        index = 0  # edge_list.currentIndex()

        for elem in selected_elements:
            if self.update_element(elem):
                index = edge_list.currentIndex()
                edge_list.setCurrentIndex(index + 1)
        self.sort_elements()
        edge_list.setCurrentIndex(index)
        self.update()

    def plot(self, event=None):
        self.energy_scale = self.dataset._axes[self.spec_dim].values
        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            spectrum = self.dataset.view.get_spectrum()
            self.axis = self.dataset.view.axes[1]
        else:
            spectrum = np.array(self.dataset)
            self.axis = self.dataset.view.axis

        if self.ui.select10.isChecked():
            if 'experiment' in self.dataset.metadata:
                exp = self.dataset.metadata['experiment']
                if 'convergence_angle' not in exp:
                    raise ValueError('need a convergence_angle in experiment of metadata dictionary ')
                alpha = exp['convergence_angle']
                beta = exp['collection_angle']
                beam_kv = exp['acceleration_voltage']

                eff_beta = eels.effective_collection_angle(self.energy_scale, alpha, beta, beam_kv)
                edges = eels.make_cross_sections(self.edges, np.array(self.energy_scale), beam_kv, eff_beta)
                self.edges = eels.fit_edges2(spectrum, self.energy_scale, edges)
                areal_density = []
                elements = []
                for key in edges:
                    if key.isdigit():  # only edges have numbers in that dictionary
                        elements.append(edges[key]['element'])
                        areal_density.append(edges[key]['areal_density'])
                areal_density = np.array(areal_density)
                out_string = '\nRelative composition: \n'
                for i, element in enumerate(elements):
                    out_string += f'{element}: {areal_density[i] / areal_density.sum() * 100:.1f}%  '

                self.model = self.edges['model']['spectrum']
                self.update()

        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.axis.clear()

        line1, = self.axis.plot(self.energy_scale, spectrum, label='spectrum')
        lines = [line1]

        def onpick(event):
            # on the pick event, find the orig line corresponding to the
            # legend proxy line, and toggle the visibility
            leg_line = event.artist
            orig_line = lined[legline]
            vis = not origline.get_visible()
            orig_line.set_visible(vis)
            # Change the alpha on the line in the legend so we can see what lines
            # have been toggled
            if vis:
                leg_line.set_alpha(1.0)
            else:
                leg_line.set_alpha(0.2)
            self.figure.canvas.draw()

        if len(self.model) > 1:
            line2, = self.axis.plot(self.energy_scale, self.model, label='model')
            line3, = self.axis.plot(self.energy_scale, spectrum - self.model, label='difference')
            line4, = self.axis.plot(self.energy_scale, (spectrum - self.model) / np.sqrt(spectrum), label='Poisson')
            lines = [line1, line2, line3, line4]
            lined = dict()

            legend = self.axis.legend(loc='upper right', fancybox=True, shadow=True)

            legend.get_frame().set_alpha(0.4)
            for legline, origline in zip(legend.get_lines(), lines):
                legline.set_picker(5)  # 5 pts tolerance
                lined[legline] = origline
            self.figure.canvas.mpl_connect('pick_event', onpick)
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)

        if self.ui.show_edges.isChecked():
            self.show_edges()
        if self.show_regions:
            self.plot_regions()
        self.figure.canvas.draw_idle()

    def plot_regions(self):
        y_min, y_max = self.axis.get_ylim()
        height = y_max - y_min

        rect = []
        if 'fit_area' in self.edges:
            color = 'blue'
            alpha = 0.2
            x_min = self.edges['fit_area']['fit_start']
            width = self.edges['fit_area']['fit_end'] - x_min
            rect.append(patches.Rectangle((x_min, y_min), width, height,
                                          edgecolor=color, alpha=alpha, facecolor=color))
            self.axis.add_patch(rect[0])
            self.axis.text(x_min, y_max, 'fit region', verticalalignment='top')
        color = 'red'
        alpha = 0.5
        for key in self.edges:
            if key.isdigit():
                x_min = self.edges[key]['start_exclude']
                width = self.edges[key]['end_exclude']-x_min
                rect.append(patches.Rectangle((x_min, y_min), width, height,
                                              edgecolor=color, alpha=alpha, facecolor=color))
                self.axis.add_patch(rect[-1])
                self.axis.text(x_min, y_max, f"exclude\n edge {int(key)+1}", verticalalignment='top')

    def show_edges(self):
        x_min, x_max = self.axis.get_xlim()
        y_min, y_max = self.axis.get_ylim()

        for key, edge in self.edges.items():
            i = 0
            if key.isdigit():
                element = edge['element']
                for sym in edge['all_edges']:
                    x = edge['all_edges'][sym]['onset'] + edge['chemical_shift']
                    if x_min < x < x_max:
                        self.axis.text(x, y_max, '\n' * i + f"{element}-{sym}",
                                       verticalalignment='top', color='black')
                        self.axis.axvline(x, ymin=0, ymax=1, color='gray')
                        i += 1

    def check_area_consistency(self):
        if self.dataset is None:
            return
        onset = float(self.ui.edit6.displayText())
        excl_start = float(self.ui.edit7.displayText())
        excl_end = float(self.ui.edit8.displayText())
        if onset < self.energy_scale[2]:
            onset = self.energy_scale[2]
            excl_start = self.energy_scale[2]
        if onset > self.energy_scale[-2]:
            onset = self.energy_scale[-2]
            excl_end = self.energy_scale[-2]
        if excl_start > onset:
            excl_start = onset
        if excl_end < onset:
            excl_end = onset

        index = self.ui.list3.currentIndex()
        self.edges[str(index)]['chemical_shift'] = onset - self.edges[str(index)]['original_onset']
        self.edges[str(index)]['onset'] = onset
        self.edges[str(index)]['end_exclude'] = excl_end
        self.edges[str(index)]['start_exclude'] = excl_start

        self.update()

    def on_list_enter(self):
        sender = self.sender()

        if sender.objectName() == 'edge_list':
            index = self.ui.list3.currentIndex()

            number_of_edges = 0
            for key in self.edges:
                if key.isdigit():
                    if int(key) > number_of_edges:
                        number_of_edges = int(key)
            number_of_edges += 1
            if index > number_of_edges:
                index = number_of_edges
            self.ui.list3.setCurrentIndex(index)
            if str(index) not in self.edges:
                self.edges[str(index)] = {'Z': 0, 'symmetry': 'K1', 'element': 'H', 'onset': 0, 'end_exclude': 0,
                                          'start_exclude': 0, 'areal_density': 0}

            self.update()
        elif sender.objectName() == 'symmetry_list':
            sym = self.ui.list5.currentText()
            index = self.ui.list3.currentIndex()
            zz = self.edges[str(index)]['Z']
            if zz > 1:
                x_section = eels.get_x_sections(zz)
                if sym in x_section:
                    start_exclude = x_section[sym]['onset'] - x_section[sym]['excl before']
                    end_exclude = x_section[sym]['onset'] + x_section[sym]['excl after']
                    self.edges[str(index)].update({'symmetry': sym, 'onset': x_section[sym]['onset'],
                                                   'end_exclude': end_exclude, 'start_exclude': start_exclude})
                    self.edges[str(index)]['chemical_shift'] = 0.0
                    self.edges[str(index)]['areal_density'] = 0.0
                    self.edges[str(index)]['original_onset'] = self.edges[index]['onset']
                    self.update()
        elif sender.objectName() == 'symmetry_method':
            self.ui.select5.setCurrentIndex(0)

    def on_check(self):
        sender = self.sender()

        if sender.objectName() == 'edge_check':
            self.show_regions = sender.isChecked()
        self.plot()

    def do_all_button_click(self):

        if self.dataset.data_type.name != 'SPECTRAL_IMAGE':
            self.do_fit_button_click()
            return

        if 'experiment' in self.dataset.metadata:
            exp = self.dataset.metadata['experiment']
            if 'convergence_angle' not in exp:
                raise ValueError('need a convergence_angle in experiment of metadata dictionary ')
            alpha = exp['convergence_angle']
            beta = exp['collection_angle']
            beam_kv = exp['acceleration_voltage']
        else:
            raise ValueError('need a experiment parameter in metadata dictionary')

        self.energy_scale = self.dataset._axes[self.spec_dim].values
        eff_beta = eels.effective_collection_angle(self.energy_scale, alpha, beta, beam_kv)
        edges = eels.make_cross_sections(self.edges, np.array(self.energy_scale), beam_kv, eff_beta)

        view = self.dataset.view
        bin_x = view.bin_x
        bin_y = view.bin_y

        start_x = view.x
        start_y = view.y

        number_of_edges = 0
        for key in self.edges:
            if key.isdigit():
                number_of_edges += 1


        results = np.zeros([int(self.dataset.shape[0]/bin_x), int(self.dataset.shape[1]/bin_y), number_of_edges])
        total_spec = int(self.dataset.shape[0]/bin_x)*int(self.dataset.shape[1]/bin_y)
        self.ui.progress.setMaximum(total_spec)
        self.ui.progress.setValue(0)
        ind = 0
        for x in range (int(self.dataset.shape[0]/bin_x)):

            for y in range(int(self.dataset.shape[1]/bin_y)):
                ind +=1
                self.ui.progress.setValue(ind)
                view.x = x*bin_x
                view.y = y*bin_y
                spectrum = view.get_spectrum()

                edges = eels.fit_edges2(spectrum, self.energy_scale, edges)
                for key, edge in edges.items():
                    if key.isdigit():
                        # element.append(edge['element'])
                        results[x,y, int(key)] = edge['areal_density']
        edges['spectrum_image_quantification'] = results
        self.ui.progress.setValue(total_spec)
        view.x = start_x
        view.y = start_y


    def do_fit_button_click(self):
        if 'experiment' in self.dataset.metadata:
            exp = self.dataset.metadata['experiment']
            if 'convergence_angle' not in exp:
                raise ValueError('need a convergence_angle in experiment of metadata dictionary ')
            alpha = exp['convergence_angle']
            beta = exp['collection_angle']
            beam_kv = exp['acceleration_voltage']

        else:
            raise ValueError('need a experiment parameter in metadata dictionary')
        self.energy_scale = self.dataset._axes[self.spec_dim].values
        eff_beta = eels.effective_collection_angle(self.energy_scale, alpha, beta, beam_kv)

        edges = eels.make_cross_sections(self.edges, np.array(self.energy_scale), beam_kv, eff_beta)

        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            spectrum = self.dataset.view.get_spectrum()
        else:
            spectrum = self.dataset
        self.edges = eels.fit_edges2(spectrum, self.energy_scale, edges)
        areal_density = []
        elements = []
        for key in edges:
            if key.isdigit():  # only edges have numbers in that dictionary
                elements.append(edges[key]['element'])
                areal_density.append(edges[key]['areal_density'])
        areal_density = np.array(areal_density)
        out_string = '\nRelative composition: \n'
        for i, element in enumerate(elements):
            out_string += f'{element}: {areal_density[i] / areal_density.sum() * 100:.1f}%  '

        self.model = self.edges['model']['spectrum']
        self.update()
        self.plot()

    def set_action(self):
        self.ui.edit1.editingFinished.connect(self.on_enter)
        self.ui.edit2.editingFinished.connect(self.on_enter)
        self.ui.list3.activated[str].connect(self.on_list_enter)
        self.ui.check3.clicked.connect(self.on_check)
        self.ui.edit4.editingFinished.connect(self.on_enter)
        self.ui.list5.activated[str].connect(self.on_list_enter)
        self.ui.select5.activated[str].connect(self.on_list_enter)

        self.ui.edit6.editingFinished.connect(self.on_enter)
        self.ui.edit7.editingFinished.connect(self.on_enter)
        self.ui.edit8.editingFinished.connect(self.on_enter)
        self.ui.edit9.editingFinished.connect(self.on_enter)

        self.ui.check10.clicked.connect(self.on_check)
        self.ui.select10.clicked.connect(self.on_check)
        self.ui.show_edges.clicked.connect(self.on_check)

        self.ui.do_all_button.clicked.connect(self.do_all_button_click)
        self.ui.do_fit_button.clicked.connect(self.do_fit_button_click)


class CurveVisualizer(object):
    """Plots a sidpy.Dataset with spectral dimension-type

    """
    def __init__(self, dset, spectrum_number=None, axis=None, leg=None, **kwargs):
        if not isinstance(dset, sidpy.Dataset):
            raise TypeError('dset should be a sidpy.Dataset object')
        if axis is None:
            self.fig = plt.figure()
            self.axis = self.fig.add_subplot(1, 1, 1)
        else:
            self.axis = axis
            self.fig = axis.figure

        self.dset = dset
        self.selection = []
        [self.spec_dim, self.energy_scale] = ft.get_dimensions_by_type('spectral', self.dset)[0]

        self.lined = dict()
        self.plot(**kwargs)

    def plot(self, **kwargs):
        if self.dset.data_type.name == 'IMAGE_STACK':
            line1, = self.axis.plot(self.energy_scale.values, self.dset[0, 0], label='spectrum', **kwargs)
        else:
            line1, = self.axis.plot(self.energy_scale.values, self.dset, label='spectrum', **kwargs)
        lines = [line1]
        if 'add2plot' in self.dset.metadata:
            data = self.dset.metadata['add2plot']
            for key, line in data.items():
                line_add, = self.axis.plot(self.energy_scale.values,  line['data'], label=line['legend'])
                lines.append(line_add)

            legend = self.axis.legend(loc='upper right', fancybox=True, shadow=True)
            legend.get_frame().set_alpha(0.4)

            for legline, origline in zip(legend.get_lines(), lines):
                legline.set_picker(True)
                legline.set_pickradius(5)  # 5 pts tolerance
                self.lined[legline] = origline
            self.fig.canvas.mpl_connect('pick_event', self.onpick)

        self.axis.axhline(0, color='gray', alpha=0.6)
        self.axis.set_xlabel(self.dset.labels[0])
        self.axis.set_ylabel(self.dset.data_descriptor)
        self.axis.ticklabel_format(style='sci', scilimits=(-2, 3))
        self.fig.canvas.draw_idle()

    def update(self, **kwargs):
        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.axis.clear()
        self.plot(**kwargs)
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)

    def onpick(self, event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        origline = self.lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legline.set_alpha(1.0)
        else:
            legline.set_alpha(0.2)
        self.fig.canvas.draw()



"""
    EELS Input Dialog for ELNES Analysis
"""
from PyQt5 import QtCore,  QtWidgets

import numpy as np
import scipy
import scipy.optimize
import scipy.signal

import sidpy
import pyTEMlib.file_tools as ft
import pyTEMlib.eels_tools as eels
import pyTEMlib.peak_dlg as peak_dlg

advanced_present = True
try:
    import advanced_eels_tools
    print('advanced EELS features enabled')
except ModuleNotFoundError:
    advanced_present = False

_version = .001


class PeakFitDialog(QtWidgets.QDialog):
    """
    EELS Input Dialog for ELNES Analysis
    """

    def __init__(self, dataset=None):
        super().__init__(None, QtCore.Qt.WindowStaysOnTopHint)
        # Create an instance of the GUI
        self.ui = peak_dlg.UiDialog(self)

        self.set_action()

        self.dataset = dataset
        self.energy_scale = np.array([])
        self.peak_out_list = []
        self.p_out = []
        self.axis = None
        self.show_regions = False
        self.show()

        if dataset is None:
            # make a dummy dataset
            dataset = ft.make_dummy_dataset('spectrum')

        if not isinstance(dataset, sidpy.Dataset):
            raise TypeError('dataset has to be a sidpy dataset')
        self.dataset = dataset
        self.spec_dim = ft.get_dimensions_by_type('spectral', dataset)
        if len(self.spec_dim) != 1:
            raise TypeError('We need exactly one SPECTRAL dimension')
        self.spec_dim = self.spec_dim[0]
        self.energy_scale = self.spec_dim[1].values.copy()

        if 'peak_fit' not in self.dataset.metadata:
            self.dataset.metadata['peak_fit'] = {}
            if 'edges' in self.dataset.metadata:
                if 'fit_area' in self.dataset.metadata['edges']:
                    self.dataset.metadata['peak_fit']['fit_start'] = \
                        self.dataset.metadata['edges']['fit_area']['fit_start']
                    self.dataset.metadata['peak_fit']['fit_end'] = self.dataset.metadata['edges']['fit_area']['fit_end']
                self.dataset.metadata['peak_fit']['peaks'] = {'0': {'position': self.energy_scale[1],
                                                                    'amplitude': 1000.0, 'width': 1.0,
                                                                    'type': 'Gauss', 'asymmetry': 0}}

        self.peaks = self.dataset.metadata['peak_fit']
        if 'fit_start' not in self.peaks:
            self.peaks['fit_start'] = self.energy_scale[1]
            self.peaks['fit_end'] = self.energy_scale[-2]

        if 'peak_model' in self.peaks:
            self.peak_model = self.peaks['peak_model']
            self.model = self.peak_model
            if 'edge_model' in self.peaks:
                self.model = self.model + self.peaks['edge_model']
        else:
            self.model = np.array([])
            self.peak_model = np.array([])
        if 'peak_out_list' in self.peaks:
            self.peak_out_list = self.peaks['peak_out_list']
        self.set_peak_list()

        self.update()
        self.dataset.plot()

        if self.dataset.data_type.name == 'SPECTRAL_IMAGE':
            if 'SI_bin_x' not in self.dataset.metadata['experiment']:
                self.dataset.metadata['experiment']['SI_bin_x'] = 1
                self.dataset.metadata['experiment']['SI_bin_y'] = 1
            bin_x = self.dataset.metadata['experiment']['SI_bin_x']
            bin_y = self.dataset.metadata['experiment']['SI_bin_y']

            self.dataset.view.set_bin([bin_x, bin_y])

        if hasattr(self.dataset.view, 'axes'):
            self.axis = self.dataset.view.axes[-1]
        elif hasattr(self.dataset.view, 'axis'):
            self.axis = self.dataset.view.axis
        self.figure = self.axis.figure

        if not advanced_present:
            self.ui.iteration_list = ['0']
            self.ui.smooth_list.clear()
            self.ui.smooth_list.addItems(self.ui.iteration_list)
            self.ui.smooth_list.setCurrentIndex(0)

        self.figure.canvas.mpl_connect('button_press_event', self.plot)
        self.plot()

    def update(self):
        # self.setWindowTitle('update')
        self.ui.edit1.setText(f"{self.peaks['fit_start']:.2f}")
        self.ui.edit2.setText(f"{self.peaks['fit_end']:.2f}")

        peak_index = self.ui.list3.currentIndex()
        if str(peak_index) not in self.peaks['peaks']:
            self.peaks['peaks'][str(peak_index)] = {'position': self.energy_scale[1], 'amplitude': 1000.0,
                                                    'width': 1.0, 'type': 'Gauss', 'asymmetry': 0}
            self.ui.list4.setCurrentText(self.peaks['peaks'][str(peak_index)]['type'])
        if 'associated_edge' in self.peaks['peaks'][str(peak_index)]:
            self.ui.unit3.setText(self.peaks['peaks'][str(peak_index)]['associated_edge'])
        else:
            self.ui.unit3.setText('')
        self.ui.edit5.setText(f"{self.peaks['peaks'][str(peak_index)]['position']:.2f}")
        self.ui.edit6.setText(f"{self.peaks['peaks'][str(peak_index)]['amplitude']:.2f}")
        self.ui.edit7.setText(f"{self.peaks['peaks'][str(peak_index)]['width']:.2f}")
        if 'asymmetry' not in self.peaks['peaks'][str(peak_index)]:
            self.peaks['peaks'][str(peak_index)]['asymmetry'] = 0.
        self.ui.edit8.setText(f"{self.peaks['peaks'][str(peak_index)]['asymmetry']:.2f}")

    def plot(self):
        spec_dim = ft.get_dimensions_by_type(sidpy.DimensionType.SPECTRAL, self.dataset)
        spec_dim = spec_dim[0]
        self.energy_scale = spec_dim[1].values
        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            spectrum = self.dataset.view.get_spectrum()
            self.axis = self.dataset.view.axes[1]
        else:
            spectrum = np.array(self.dataset)
            self.axis = self.dataset.view.axis

        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.axis.clear()

        self.axis.plot(self.energy_scale, spectrum, label='spectrum')
        if len(self.model) > 1:
            self.axis.plot(self.energy_scale, self.model, label='model')
            self.axis.plot(self.energy_scale, spectrum - self.model, label='difference')
            self.axis.plot(self.energy_scale, (spectrum - self.model) / np.sqrt(spectrum), label='Poisson')
            self.axis.legend()
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)
        self.axis.figure.canvas.draw_idle()

        for index, peak in self.peaks['peaks'].items():
            p = [peak['position'], peak['amplitude'], peak['width']]
            self.axis.plot(self.energy_scale, eels.gauss(self.energy_scale, p))

    def fit_peaks(self):
        """Fit spectrum with peaks given in peaks dictionary"""
        p_in = []
        for key, peak in self.peaks['peaks'].items():
            if key.isdigit():
                p_in.append(peak['position'])
                p_in.append(peak['amplitude'])
                p_in.append(peak['width'])
        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            spectrum = self.dataset.view.get_spectrum()
        else:
            spectrum = np.array(self.dataset)
        energy_scale = np.array(self.energy_scale)
        start_channel = np.searchsorted(energy_scale, self.peaks['fit_start'])
        end_channel = np.searchsorted(energy_scale, self.peaks['fit_end'])

        energy_scale = self.energy_scale[start_channel:end_channel]
        if self.energy_scale[0] > 0:
            if 'edges' not in self.dataset.metadata:
                return
            if 'model' not in self.dataset.metadata['edges']:
                return
            model = self.dataset.metadata['edges']['model']['spectrum'][start_channel:end_channel]

        else:
            model = np.zeros(end_channel - start_channel)

        difference = np.array(spectrum[start_channel:end_channel] - model)

        [self.p_out, _] = scipy.optimize.leastsq(eels.residuals_smooth, np.array(p_in), ftol=1e-3,
                                                 args=(energy_scale, difference, False))
        self.peak_model = np.zeros(len(self.energy_scale))
        fit = eels.model_smooth(energy_scale, self.p_out, False)
        self.peak_model[start_channel:end_channel] = fit
        if self.energy_scale[0] > 0:
            self.model = self.dataset.metadata['edges']['model']['spectrum']
        else:
            self.model = np.zeros(len(self.energy_scale))

        self.dataset.metadata['peak_fit']['edge_model'] = self.model
        self.model = self.model + self.peak_model
        self.dataset.metadata['peak_fit']['peak_model'] = self.peak_model

        for key, peak in self.peaks['peaks'].items():
            if key.isdigit():
                p_index = int(key)*3
                self.peaks['peaks'][key] = {'position': self.p_out[p_index],
                                            'amplitude': self.p_out[p_index+1],
                                            'width': self.p_out[p_index+2],
                                            'associated_edge': ''}

        self.find_associated_edges()
        self.find_white_lines()
        self.update()
        self.plot()

    def smooth(self):
        """Fit lots of Gaussian to spectrum and let the program sort it out

        We sort the peaks by area under the Gaussians, assuming that small areas mean noise.

        """
        iterations = int(self.ui.smooth_list.currentIndex())

        self.peak_model, self.peak_out_list, number_of_peaks = smooth(self.dataset, iterations, advanced_present)

        spec_dim = ft.get_dimensions_by_type('SPECTRAL', self.dataset)[0]
        if spec_dim[1][0] > 0:
            self.model = self.dataset.metadata['edges']['model']['spectrum']
        else:
            self.model = np.zeros(len(spec_dim[1]))

        self.ui.find_edit.setText(str(number_of_peaks))

        self.dataset.metadata['peak_fit']['edge_model'] = self.model
        self.model = self.model + self.peak_model
        self.dataset.metadata['peak_fit']['peak_model'] = self.peak_model
        self.dataset.metadata['peak_fit']['peak_out_list'] = self.peak_out_list

        self.update()
        self.plot()

    def find_associated_edges(self):
        onsets = []
        edges = []
        if 'edges' in self.dataset.metadata:
            for key, edge in self.dataset.metadata['edges'].items():
                if key.isdigit():
                    element = edge['element']
                    for sym in edge['all_edges']:  # TODO: Could be replaced with exclude
                        onsets.append(edge['all_edges'][sym]['onset'] + edge['chemical_shift'])
                        # if 'sym' == edge['symmetry']:
                        edges.append([key, f"{element}-{sym}", onsets[-1]])
        for key, peak in self.peaks['peaks'].items():
            if key.isdigit():
                distance = self.energy_scale[-1]
                index = -1
                for ii, onset in enumerate(onsets):
                    if onset < peak['position'] < onset+50:
                        if distance > np.abs(peak['position'] - onset):
                            distance = np.abs(peak['position'] - onset)  # TODO: check whether absolute is good
                            distance_onset = peak['position'] - onset
                            index = ii
                if index >= 0:
                    peak['associated_edge'] = edges[index][1]  # check if more info is necessary
                    peak['distance_to_onset'] = distance_onset

    def find_white_lines(self):
        white_lines = {}
        for index, peak in self.peaks['peaks'].items():
            if index.isdigit():
                if 'associated_edge' in peak:
                    if peak['associated_edge'][-2:] in ['L3', 'L2', 'M5', 'M4']:
                        if peak['distance_to_onset'] < 10:
                            area = np.sqrt(2 * np.pi) * peak['amplitude'] * np.abs(peak['width']/np.sqrt(2 * np.log(2)))
                            if peak['associated_edge'] not in white_lines:
                                white_lines[peak['associated_edge']] = 0.
                            if area > 0:
                                white_lines[peak['associated_edge']] += area  # TODO: only positive ones?
        white_line_ratios = {}
        white_line_sum = {}
        for sym, area in white_lines.items():
            if sym[-2:] in ['L2', 'M4', 'M2']:
                if area > 0 and f"{sym[:-1]}{int(sym[-1]) + 1}" in white_lines:
                    if white_lines[f"{sym[:-1]}{int(sym[-1]) + 1}"] > 0:
                        white_line_ratios[f"{sym}/{sym[-2]}{int(sym[-1]) + 1}"] = area / white_lines[
                            f"{sym[:-1]}{int(sym[-1]) + 1}"]
                        white_line_sum[f"{sym}+{sym[-2]}{int(sym[-1]) + 1}"] = (
                                    area + white_lines[f"{sym[:-1]}{int(sym[-1]) + 1}"])

                        areal_density = 1.
                        if 'edges' in self.dataset.metadata:
                            for key, edge in self.dataset.metadata['edges'].items():
                                if key.isdigit():
                                    if edge['element'] == sym.split('-')[0]:
                                        areal_density = edge['areal_density']
                                        break
                        white_line_sum[f"{sym}+{sym[-2]}{int(sym[-1]) + 1}"] /= areal_density

        self.peaks['white_lines'] = white_lines
        self.peaks['white_line_ratios'] = white_line_ratios
        self.peaks['white_line_sums'] = white_line_sum
        self.ui.wl_list = []
        self.ui.wls_list = []
        if len(self.peaks['white_line_ratios']) > 0:
            for key in self.peaks['white_line_ratios']:
                self.ui.wl_list.append(key)
            for key in self.peaks['white_line_sums']:
                self.ui.wls_list.append(key)

            self.ui.listwl.clear()
            self.ui.listwl.addItems(self.ui.wl_list)
            self.ui.listwl.setCurrentIndex(0)
            self.ui.unitswl.setText(f"{self.peaks['white_line_ratios'][self.ui.wl_list[0]]:.2f}")

            self.ui.listwls.clear()
            self.ui.listwls.addItems(self.ui.wls_list)
            self.ui.listwls.setCurrentIndex(0)
            self.ui.unitswls.setText(f"{self.peaks['white_line_sums'][self.ui.wls_list[0]]*1e6:.4f} ppm")
        else:
            self.ui.wl_list.append('Ratio')
            self.ui.wls_list.append('Sum')

            self.ui.listwl.clear()
            self.ui.listwl.addItems(self.ui.wl_list)
            self.ui.listwl.setCurrentIndex(0)
            self.ui.unitswl.setText('')

            self.ui.listwls.clear()
            self.ui.listwls.addItems(self.ui.wls_list)
            self.ui.listwls.setCurrentIndex(0)
            self.ui.unitswls.setText('')

    def find_peaks(self):
        number_of_peaks = int(str(self.ui.find_edit.displayText()).strip())

        # is now sorted in smooth function
        # flat_list = [item for sublist in self.peak_out_list for item in sublist]
        # new_list = np.reshape(flat_list, [len(flat_list) // 3, 3])
        # arg_list = np.argsort(np.abs(new_list[:, 1]))

        self.ui.peak_list = []
        self.peaks['peaks'] = {}
        for i in range(number_of_peaks):
            self.ui.peak_list.append(f'Peak {i+1}')
            p = self.peak_out_list[i]
            self.peaks['peaks'][str(i)] = {'position': p[0], 'amplitude': p[1], 'width': p[2], 'type': 'Gauss',
                                           'asymmetry': 0}

        self.ui.peak_list.append(f'add peak')
        self.ui.list3.clear()
        self.ui.list3.addItems(self.ui.peak_list)
        self.ui.list3.setCurrentIndex(0)
        self.find_associated_edges()
        self.find_white_lines()

        self.update()
        self.plot()

    def set_peak_list(self):
        self.ui.peak_list = []
        if 'peaks' not in self.peaks:
            self.peaks['peaks'] = {}
        key = 0
        for key in self.peaks['peaks']:
            if key.isdigit():
                self.ui.peak_list.append(f'Peak {int(key) + 1}')
        self.ui.find_edit.setText(str(int(key) + 1))
        self.ui.peak_list.append(f'add peak')
        self.ui.list3.clear()
        self.ui.list3.addItems(self.ui.peak_list)
        self.ui.list3.setCurrentIndex(0)

    def on_enter(self):
        if self.sender() == self.ui.edit1:
            value = float(str(self.ui.edit1.displayText()).strip())
            if value < self.energy_scale[0]:
                value = self.energy_scale[0]
            if value > self.energy_scale[-5]:
                value = self.energy_scale[-5]
            self.peaks['fit_start'] = value
            self.ui.edit1.setText(str(self.peaks['fit_start']))
        elif self.sender() == self.ui.edit2:
            value = float(str(self.ui.edit2.displayText()).strip())
            if value < self.energy_scale[5]:
                value = self.energy_scale[5]
            if value > self.energy_scale[-1]:
                value = self.energy_scale[-1]
            self.peaks['fit_end'] = value
            self.ui.edit2.setText(str(self.peaks['fit_end']))
        elif self.sender() == self.ui.edit5:
            value = float(str(self.ui.edit5.displayText()).strip())
            peak_index = self.ui.list3.currentIndex()
            self.peaks['peaks'][str(peak_index)]['position'] = value
        elif self.sender() == self.ui.edit6:
            value = float(str(self.ui.edit6.displayText()).strip())
            peak_index = self.ui.list3.currentIndex()
            self.peaks['peaks'][str(peak_index)]['amplitude'] = value
        elif self.sender() == self.ui.edit7:
            value = float(str(self.ui.edit7.displayText()).strip())
            peak_index = self.ui.list3.currentIndex()
            self.peaks['peaks'][str(peak_index)]['width'] = value

    def on_list_enter(self):
        # self.setWindowTitle('list')
        if self.sender() == self.ui.list3:
            if self.ui.list3.currentText().lower() == 'add peak':
                peak_index = self.ui.list3.currentIndex()
                self.ui.list3.insertItem(peak_index, f'Peak {peak_index+1}')
                self.peaks['peaks'][str(peak_index+1)] = {'position': self.energy_scale[1],
                                                          'amplitude': 1000.0, 'width': 1.0,
                                                          'type': 'Gauss', 'asymmetry': 0}
                self.ui.list3.setCurrentIndex(peak_index)
            self.update()

        elif self.sender() == self.ui.listwls or self.sender() == self.ui.listwl:
            wl_index = self.sender().currentIndex()

            self.ui.listwl.setCurrentIndex(wl_index)
            self.ui.unitswl.setText(f"{self.peaks['white_line_ratios'][self.ui.wl_list[wl_index]]:.2f}")
            self.ui.listwls.setCurrentIndex(wl_index)
            self.ui.unitswls.setText(f"{self.peaks['white_line_sums'][self.ui.wls_list[wl_index]] * 1e6:.4f} ppm")

    def set_action(self):
        pass
        self.ui.edit1.editingFinished.connect(self.on_enter)
        self.ui.edit2.editingFinished.connect(self.on_enter)
        self.ui.edit5.editingFinished.connect(self.on_enter)
        self.ui.edit6.editingFinished.connect(self.on_enter)
        self.ui.edit7.editingFinished.connect(self.on_enter)
        self.ui.edit8.editingFinished.connect(self.on_enter)
        self.ui.list3.activated[str].connect(self.on_list_enter)
        self.ui.find_button.clicked.connect(self.find_peaks)
        self.ui.smooth_button.clicked.connect(self.smooth)
        self.ui.fit_button.clicked.connect(self.fit_peaks)
        self.ui.listwls.activated[str].connect(self.on_list_enter)
        self.ui.listwl.activated[str].connect(self.on_list_enter)

def smooth(dataset, iterations, advanced_present):
    """Gaussian mixture model (non-Bayesian)

    Fit lots of Gaussian to spectrum and let the program sort it out
    We sort the peaks by area under the Gaussians, assuming that small areas mean noise.

    """

    # TODO: add sensitivity to dialog and the two functions below
    peaks = dataset.metadata['peak_fit']

    if advanced_present and iterations > 1:
        peak_model, peak_out_list = advanced_eels_tools.smooth(dataset, peaks['fit_start'],
                                                               peaks['fit_end'], iterations=iterations)
    else:
        peak_model, peak_out_list = eels.find_peaks(dataset, peaks['fit_start'], peaks['fit_end'])
        peak_out_list = [peak_out_list]

    flat_list = [item for sublist in peak_out_list for item in sublist]
    new_list = np.reshape(flat_list, [len(flat_list) // 3, 3])
    area = np.sqrt(2 * np.pi) * np.abs(new_list[:, 1]) * np.abs(new_list[:, 2] / np.sqrt(2 * np.log(2)))
    arg_list = np.argsort(area)[::-1]
    area = area[arg_list]
    peak_out_list = new_list[arg_list]

    number_of_peaks = np.searchsorted(area * -1, -np.average(area))

    return peak_model, peak_out_list, number_of_peaks


"""utility functions for sidpy; will move to sidpy"""
import numpy as np
import sidpy
import h5py
import pyNSID
import os
import ipywidgets as widgets
from IPython.display import display


class ChooseDataset(object):
    """Widget to select dataset object """

    def __init__(self, input_object, show_dialog=True):
        if isinstance(input_object, sidpy.Dataset):
            if isinstance(input_object.h5_dataset, h5py.Dataset):
                self.current_channel = input_object.h5_dataset.parent
        elif isinstance(input_object, h5py.Group):
            self.current_channel = input_object
        elif isinstance(input_object, h5py.Dataset):
            self.current_channel = input_object.parent
        else:
            raise ValueError('Need hdf5 group or sidpy Dataset to determine image choices')
        self.dataset_names = []
        self.dataset_list = []
        self.dataset_type = None
        self.dataset = None
        self.reader = pyNSID.NSIDReader(self.current_channel.file.filename)

        self.get_dataset_list()
        self.select_image = widgets.Dropdown(options=self.dataset_names,
                                             value=self.dataset_names[0],
                                             description='Select image:',
                                             disabled=False,
                                             button_style='')
        if show_dialog:
            display(self.select_image)

        self.select_image.observe(self.set_dataset, names='value')
        self.set_dataset(0)
        self.select_image.index = (len(self.dataset_names) - 1)

    def get_dataset_list(self):
        """ Get by Log number sorted list of datasets"""
        datasets = self.reader.read()
        order = []
        for dset in datasets:
            if self.dataset_type is None or dset.data_type == self.data_type:
                if 'Log' in dset.title:
                    position = dset.title.find('Log_') + 4
                    order.append(int(dset.title[position:position + 3])+1)
                else:
                    order.append(0)
        for index in np.argsort(order):
            dset = datasets[index]
            self.dataset_names.append('/'.join(dset.title.replace('-', '_').split('/')[-1:]))
            self.dataset_list.append(dset)

    def set_dataset(self, b):
        index = self.select_image.index
        self.dataset = self.dataset_list[index]


def get_dimensions_by_order(dims_in, dataset):
    """get dimension

    Parameters
    ----------
    dims_in: int or list of int
        the dimensions by numerical order
    dataset: sidpy.Dataset

    Returns
    -------
    dims_out: list of dimensions
    """

    if isinstance(dims_in, int):
        dims_in = [dims_in]
    dims_out = []
    for item in dims_in:
        if isinstance(item, int):
            if item in dataset._axes:
                dims_out.append([item, dataset._axes[item]])
    return dims_out


def get_dimensions_by_type(dims_in, dataset):
    """ get dimension by dimension_type name

    Parameters
    ----------
    dims_in: dimension_type or list of dimension_types
        the dimensions by numerical order
    dataset: sidpy.Dataset

    Returns
    -------
    dims_out: list of dimensions
    """

    if isinstance(dims_in, (str, sidpy.DimensionType)):
        dims_in = [dims_in]
    for i in range(len(dims_in)):
        if isinstance(dims_in[i], str):
            dims_in[i] = sidpy.DimensionType[dims_in[i].upper()]
    dims_out = []
    for dim, axis in dataset._axes.items():
        if axis.dimension_type in dims_in:
            dims_out.append([dim, dataset._axes[dim]])
    return dims_out


def make_dummy_dataset(value_type):
    """Make a dummy sidpy.Dataset """

    assert isinstance(value_type, sidpy.DataType)
    if type == sidpy.DataType.SPECTRUM:
        dataset = sidpy.Dataset.from_array(np.arange(100))
        dataset.data_type = 'spectrum'
        dataset.units = 'counts'
        dataset.quantity = 'intensity'

        dataset.set_dimension(0, sidpy.Dimension(np.arange(dataset.shape[0]) + 70, name='energy_scale'))
        dataset.dim_0.dimension_type = 'spectral'
        dataset.dim_0.units = 'eV'
        dataset.dim_0.quantity = 'energy loss'
    else:
        raise NotImplementedError('not implemented')
    return dataset


def plot(dataset):
    dataset.plot()


def get_image_dims(dataset):
    """Get all spatial dimensions"""

    image_dims = []
    for dim, axis in dataset._axes.items():
        if axis.dimension_type == sidpy.DimensionType.SPATIAL:
            image_dims.append(dim)
    return image_dims


def get_extent(dataset):
    """get extent to plot with matplotlib"""
    image_dims = get_image_dims(dataset)
    return dataset.get_extent(image_dims)


class FileWidget_sidpy(object):
    """Widget to select directories or widgets from a list

    Works in google colab.

    The widgets converts the name of the nion file to the one in Nion's swift software,
    because it is otherwise incomprehensible
    Use as:
    >>from google.colab import drive
    >>drive.mount("/content/drive")
    >>file_list = pyTEMlib.FileWidget()
    next code cell:
    >>dataset = pyTEMlib.open_file(file_list.file_name)

    """

    def __init__(self, dir_name='.', extension=['*']):

        if dir_name == '.':
            self.get_directory('.')
            """elif dir_name == None:
            fp = open(config_path+'\\path.txt','r')
            path = fp.read()
            fp.close()
            self.get_directory(path)
            self.dir_name = path
            """
        elif os.path.isdir(dir_name):
            self.get_directory(dir_name)
            self.dir_name = dir_name
        else:
            self.dir_name = '.'
            self.get_directory(self.dir_name)

        self.dir_list == ['.']
        self.extensions = extension
        self.file_name = ''

        self.select_files = widgets.Select(
            options=self.dir_list,
            value=self.dir_list[0],
            description='Select file:',
            disabled=False,
            rows=10,
            layout=widgets.Layout(width='70%')
        )
        display(self.select_files)
        self.set_options()
        self.select_files.observe(self.get_file_name, names='value')

    def get_directory(self, directory=None):
        self.dir_name = directory
        self.dir_dictionary = {}
        self.dir_list = []
        self.dir_list = ['.', '..'] + os.listdir(directory)

    def set_options(self):
        self.dir_name = os.path.abspath(os.path.join(self.dir_name, self.dir_list[self.select_files.index]))
        dir_list = os.listdir(self.dir_name)

        file_list = []
        display_file_list = []
        directory_list = []

        for i in range(len(dir_list)):
            name = dir_list[i]
            full_name = os.path.join(self.dir_name, name)

            if os.path.isfile(full_name):
                size = os.path.getsize(full_name) * 2 ** -20
                basename, extension = os.path.splitext(name)
                if self.extensions[0] == 'hf5':
                    if extension in ['.hf5']:
                        file_list.append(dir_list[i])
                        display_file_list.append(f" {name}  - {size:.1f} MB")
                else:
                    file_list.append(dir_list[i])
                    # if extension in ['.h5', '.ndata']:
                    #    # tags = open_file(os.path.join(self.dir_name, name))
                    #    display_file_list.append(f" {tags['original_name']}{extension}  - {size:.1f} MB")
                    if extension in ['.hf5']:
                        display_file_list.append(f" {name}  -- {size:.1f} MB")
                    else:
                        display_file_list.append(f' {name}  - {size:.1f} MB')
            else:
                directory_list.append(name)

        sort = np.argsort(directory_list)
        self.dir_list = ['.', '..']
        self.display_list = ['.', '..']
        for j in sort:
            self.display_list.append(f' * {directory_list[j]}')
            self.dir_list.append(directory_list[j])

        sort = np.argsort(display_file_list)

        for i, j in enumerate(sort):
            if '--' in dir_list[j]:
                self.display_list.append(f' {i:3} {display_file_list[j]}')
            else:
                self.display_list.append(f' {i:3}   {display_file_list[j]}')
            self.dir_list.append(file_list[j])

        self.dir_label = os.path.split(self.dir_name)[-1] + ':'
        self.select_files.options = self.display_list

    def get_file_name(self, b):

        if os.path.isdir(os.path.join(self.dir_name, self.dir_list[self.select_files.index])):
            self.set_options()

        elif os.path.isfile(os.path.join(self.dir_name, self.dir_list[self.select_files.index])):
            self.file_name = os.path.join(self.dir_name, self.dir_list[self.select_files.index])



"""EEL edge database from EELSLab, not longer used
"""
# -*- coding: utf-8 -*-
# Copyright © 2007 Francisco Javier de la Peña
#
# This file is part of EELSLab.
#
# EELSLab is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# EELSLab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EELSLab; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA


import csv
import os

from pyTEMlib.config_dir import config_path


file_path = os.path.join(config_path, 'edges_db.csv')
f = open(file_path, 'r')
reader = csv.reader(f)
edges_dict = {}
for row in reader:
    twin_subshell = None
    element, subshell = row[0].split('.')
    Z = row[1]
    if (element in edges_dict) is not True:
        edges_dict[element] = {}
        edges_dict[element]['subshells'] = {}
        edges_dict[element]['Z'] = Z
    factor = 1.0
    if row[3] != '':
        if subshell == "L3":
            twin_subshell = "L2"
            factor = 0.5
        if subshell == "M3":
            twin_subshell = "M2"
            factor = 0.5
        if subshell == "M5":
            twin_subshell = "M4"
            factor = 4/6.
        if subshell == "N3":
            twin_subshell = "N2"
            factor = 2/4.
        if subshell == "N5":
            twin_subshell = "N4"
            factor = 4/6.
        if subshell == "N7":
            twin_subshell = "N6"
            factor = 6/8.
        if subshell == "O5":
            twin_subshell = "O4"
            factor = 4/6.

    edges_dict[element]['subshells'][subshell] = {}
    edges_dict[element]['subshells'][subshell]['onset_energy'] = float(row[2])
    edges_dict[element]['subshells'][subshell]['filename'] = row[0]
    edges_dict[element]['subshells'][subshell]['relevance'] = row[4]
    edges_dict[element]['subshells'][subshell]['factor'] = 1

    if twin_subshell is not None:
        edges_dict[element]['subshells'][twin_subshell] = {}
        edges_dict[element]['subshells'][twin_subshell]['onset_energy'] = float(row[3])
        edges_dict[element]['subshells'][twin_subshell]['filename'] = row[0]
        edges_dict[element]['subshells'][twin_subshell]['relevance'] = row[4]
        edges_dict[element]['subshells'][twin_subshell]['factor'] = factor
